/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 1997-2012 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */
// Portions Copyright [2018-2019] [Payara Foundation and/or its affiliates]

package com.sun.enterprise.admin.util;

import com.sun.enterprise.security.store.AsadminTruststore;
import com.sun.enterprise.universal.i18n.LocalStringsImpl;
import org.jline.reader.LineReader;
import org.jline.reader.LineReaderBuilder;
import org.jline.terminal.impl.DumbTerminal;

import javax.net.ssl.X509TrustManager;
import java.io.IOException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.text.DateFormat;
import java.util.Date;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * An implementation of {@link X509TrustManager} that provides basic support
 * for Trust Management.  It checks if the server is trusted and displays the
 * certificate that was received from the server.  The user is then prompted
 * to confirm the certificate.  If confirmed, the certificate is entered into
 * the client side asadmintruststore (default name is ~/.gfclient/truststore).
 * Once in the truststore, the user is never prompted to confirm a second time.
 */
public class AsadminTrustManager implements X509TrustManager {
        
    private final Object alias;    
    private boolean alreadyInvoked;
    private boolean interactive = true;
    private CertificateException lastCertException;
    private RuntimeException lastRuntimeException;
    
    private static final LocalStringsImpl STRING_MANAGER = new LocalStringsImpl(AsadminTrustManager.class);
    private static final Logger logger = Logger.getLogger(AsadminTrustManager.class.getName());

    /**
     * Creates an instance of the AsadminTrustManager
     * @param alias The toString() of the alias object concatenated with a
     * date/time stamp is used as the alias of the trusted server certificate
     * in the client side trust store.  When null,
     * only a date / timestamp is used as an alias.
     */    
    public AsadminTrustManager(Object alias, Map env) {
        this.alias = alias;
        alreadyInvoked = false;
        lastCertException = null;
        lastRuntimeException = null;
    }
    
    /**
     * Creates an instance of the SunOneBasicX509TrustManager
     * A date/time stamp is used of the trusted server certificate in the
     * client side trust store.
     */
    public AsadminTrustManager() {
        this (null, null);
    }
    
    /**
     * Set the interactive mode for the trust manager.  If false, it will
     * not prompt for any confirmations and will just trust certificates.
     * By default it is true.
     */
    public void setInteractive(boolean mode) {
        interactive = mode;
    }

    /**
     * Checks if client is trusted given the certificate chain and
     * authorization type string, e.g., "RSA".
     * @throws CertificateException
     * @throws UnsupportedOperationException
     */
    @Override
    public void checkClientTrusted(X509Certificate[] x509Certificate,
                                String authType) throws CertificateException {
        throw new UnsupportedOperationException(
                            "Not Implemented for Client Trust Management");
    }
 
    /**
     * Checs if the server is trusted.
     * @param chain The server certificate to be  validated.
     * @param authType
     * @throws CertificateException
     */    
    @Override
    public void checkServerTrusted(X509Certificate[] chain, String authType) 
        throws CertificateException {
        // The alreadyInvoked flag keeps track of whether we have already
        // prompted the user.  Unfortunately, checkServerTrusted is called
        // two times and we want to avoid prompting the user twice. I'm not
        // sure of the root cause of this problem (i.e., why it is called
        // twice).  In addition, we keep track of any exception that occurred
        // on the first invocation and propagate that back.
        if (!alreadyInvoked) {
            alreadyInvoked = true;            
            try {
                checkCertificate(chain);                
            } catch (RuntimeException ex) {
                lastRuntimeException = ex;
                throw ex;
            } catch (CertificateException ex) {
                lastCertException = ex;
                throw ex;
            } 
        } else {
            if (lastRuntimeException != null) {
                throw lastRuntimeException;
            } else if (lastCertException != null) {
                throw lastCertException;
            }
        }
    }
 
    @Override
    public X509Certificate[] getAcceptedIssuers() {
        return new X509Certificate[0];
    }
    
    /**
     * Displays the certificate and prompts the user whether or 
     * not it is trusted.
     * @param c
     * @throws IOException
     * @return true if the user trusts the certificate
     */    
    private boolean isItOKToAddCertToTrustStore(X509Certificate c) {
        if (!interactive) {
            return true;
        }

        String result = null;
        LineReader lineReader = null;
        try {
            lineReader = LineReaderBuilder.builder()
                    .terminal(new DumbTerminal(System.in, System.out))
                    .build();

            result = lineReader.readLine(STRING_MANAGER.get("certificateTrustPrompt"));
        } catch (IOException ioe) {
            logger.log(Level.WARNING, "Error instantiating console", ioe);
        }
        finally {
            if (lineReader != null && lineReader.getTerminal() != null) {
                try {
                    lineReader.getTerminal().close();
                } catch (IOException ioe) {
                    logger.log(Level.WARNING, "Error closing terminal", ioe);
                }
            }
        }
        return result != null && result.equalsIgnoreCase("y");
    }
 
    private String getAliasName() {
        String aliasName = alias != null ? alias.toString() : "";
        // We append a timestamp to the alias to ensure that it is unqiue.    
        DateFormat f =
            DateFormat.getDateTimeInstance(DateFormat.LONG, DateFormat.LONG);
        aliasName += ":" + f.format(new Date());        
        return aliasName;
    }
 
    /**
     * This function validates the cert and ensures that it is trusted.
     * @param chain
     * @throws RuntimeException
     * @throws CertificateException
     */    
    protected void checkCertificate(X509Certificate[] chain)
                                throws RuntimeException, CertificateException,
                                IllegalArgumentException {        
        if (chain == null || chain.length == 0) {
            throw new IllegalArgumentException (STRING_MANAGER.get(
                "emptyServerCertificate"));
        } 
        // First ensure that the certificate is valid.
        for (int i = 0 ; i < chain.length ; i ++) {
            chain[i].checkValidity();   
        }
        try {
            /*
             * Open the trust store, prompting the user if needed for a valid
             * password to use.
             */
            AsadminTruststore truststore = AsadminTruststore.newInstance();
            // if the certificate already exists in the truststore,
            // it is implicitly trusted
            if (!truststore.certificateExists(chain[0])) {
                // if the certificate does not exist in the truststore,
                // then we prompt the user.  Upon confirmation from the user,
                // the certificate is added to the truststore.
                if (isItOKToAddCertToTrustStore(chain[0])) {
                    truststore.addCertificate(getAliasName(), chain[0]);
                } else {
                    throw new CertificateException(STRING_MANAGER.get("serverCertificateNotTrusted"));
                }
            }     
        } catch (CertificateException ex) {
            throw ex;
        } catch (Exception e) {        
            throw new RuntimeException(e);
        }        
    }
}
